@@ -20,6 +20,7 @@ import json
 from dataclasses import dataclass
 
 import flask
+import threading
 
 # for local use & forwarding
 from flask import Response, request
@@ -98,40 +99,69 @@ class Reference:
     ```
 
     """
+    class Local(object):
+        pass
 
-    def __init__(self, obj: Any = None, set_name: str = "set"):
+    def __init__(self, obj: Any = None, set_name: str = "set", fun: Optional[Callable] = None):
         """Constructor parameters:
 
+        - set_name: provide another prefix for the "set" functions.
         - obj: object to be wrapped, can also be provided later.
-        - set_name: provide another name for the "set" function.
+        - fun: function to generated a per-thread wrapped object.
         """
-        self._obj = None
+        if obj and fun:
+            raise Exception("reference cannot have both obj and fun")
+        elif obj:
+            self._set_obj(obj)
+        elif fun:
+            self._set_fun(fun)
         if set_name:
             setattr(self, set_name, getattr(self, "_set_obj"))
-        obj and self._set_obj(obj)
+            setattr(self, set_name + "_obj", getattr(self, "_set_obj"))
+            setattr(self, set_name + "_fun", getattr(self, "_set_fun"))
 
     def _set_obj(self, obj):
-        """Set current wrapped object, possibly replacing the previous one."""
+        """Set current wrapped object."""
         log.debug(f"setting reference to {obj} ({type(obj)})")
-        self._obj = obj
+        self._fun = None
+        self._nthreads = 1
+        self._local = self.Local()
+        self._local.obj = obj
         return obj
 
+    def _set_fun(self, fun: Callable[[int], Any]):
+        """Set current wrapped object generation function."""
+        self._fun = fun
+        self._nthreads = 0
+        self._local = threading.local()
+        pass
+
+    def _get_obj(self):
+        """Get current wrapped object."""
+        if self._fun and not hasattr(self._local, "obj"):
+            self._local.obj = self._fun(self._nthreads)
+            self._nthreads += 1
+        return self._local.obj
+
     def __getattr__(self, item):
-        """Forward everything to contained object."""
-        return self._obj.__getattribute__(item)
+        """Forward everything unknown to contained object."""
+        return self._get_obj().__getattribute__(item)
 
     # also forward a few special methods
     def __str__(self):
-        return self._obj.__str__()
+        return self._get_obj().__str__()
 
     def __repr__(self):
-        return self._obj.__repr__()
+        return self._get_obj().__repr__()
 
     def __eq__(self, v):
-        return self._obj.__eq__(v)
+        return self._get_obj().__eq__(v)
 
     def __ne__(self, v):
-        return self._obj.__ne__(v)
+        return self._get_obj().__ne__(v)
+
+    def __hash__(self):
+        return self._get_obj().__hash__()
 
 
 class Flask(flask.Flask):
@@ -231,7 +261,6 @@ class FlaskSimpleAuth:
             JsonData: json.loads,
         }
         self._auth: List[str] = []
-        self._saved_auth: Optional[List[str]] = None
         self._http_auth = None
         self._pm = None
         self._cache: Optional[MutableMapping[str, str]] = None
@@ -239,6 +268,8 @@ class FlaskSimpleAuth:
         self._server_error: int = _DEFAULT_SERVER_ERROR
         self._not_found_error: int = _DEFAULT_NOT_FOUND_ERROR
         self._secure: bool = True
+        self._names: Set[str] = set()
+        self._local = threading.local()
         # actual main initialization is deferred to `init_app`
         self._initialized = False
 
@@ -262,7 +293,7 @@ class FlaskSimpleAuth:
     def _auth_has(self, *auth):
         """Tell whether current authentication includes any of these schemes."""
         for a in auth:
-            if a in self._auth:
+            if a in self._local.auth:
                 return True
         return False
 
@@ -281,13 +312,14 @@ class FlaskSimpleAuth:
 
     def _auth_reset_user(self):
         """Before request hook to cleanup authentication and authorization."""
-        self._user_set = False
-        self._user = None
-        self._need_authorization = True
+        self._local.user_set = False
+        self._local.user = None
+        self._local.need_authorization = True
+        self._local.auth = self._auth
 
     def _auth_post_check(self, res: Response):
         """After request hook to detect missing authorizations."""
-        if res.status_code < 400 and self._need_authorization:
+        if res.status_code < 400 and self._local.need_authorization:
             method, path = request.method, request.path
             if not (self._cors and method == "OPTIONS"):
                 log.error(f"missing authorization on {method} {path}")
@@ -311,7 +343,7 @@ class FlaskSimpleAuth:
         # NOTE thanks to max_age the client should not send stale cookies
         if self._carrier == "cookie":
             assert self._token and self._name
-            if self._user and self._can_create_token():
+            if self._local.user and self._can_create_token():
                 if self._name in request.cookies:
                     user, exp = self._get_any_token_auth_exp(request.cookies[self._name])
                     # renew token when closing expiration
@@ -320,7 +352,7 @@ class FlaskSimpleAuth:
                 else:
                     set_cookie = True
                 if set_cookie:
-                    res.set_cookie(self._name, self.create_token(self._user),
+                    res.set_cookie(self._name, self.create_token(self._local.user),
                                    max_age=int(60 * self._delay))
         return res
 
@@ -330,16 +362,13 @@ class FlaskSimpleAuth:
             # FIXME should it prioritize based on self._auth order?
             if self._auth_has("basic", "password"):
                 res.headers["WWW-Authenticate"] = f"Basic realm=\"{self._realm}\""
-            elif "token" in self._auth and self._carrier == "bearer":
+            elif "token" in self._local.auth and self._carrier == "bearer":
                 res.headers["WWW-Authenticate"] = f"{self._name} realm=\"{self._realm}\""
             elif self._auth_has("http-basic", "http-digest", "http-token", "digest"):
                 assert self._http_auth
                 res.headers["WWW-Authenticate"] = self._http_auth.authenticate_header()
             # else: scheme does not rely on WWW-Authenticate…
         # else: no need for WWW-Authenticate
-        # restore temporary auth if needed
-        if self._saved_auth:
-            self._auth, self._saved_auth = self._saved_auth, None
         return res
 
     def _params(self):
@@ -451,6 +480,7 @@ class FlaskSimpleAuth:
             self._auth = auth
         for a in self._auth:
             assert a in self._FSA_AUTH
+        self._local.auth = self._auth
         #
         # web apps…
         #
@@ -520,13 +550,16 @@ class FlaskSimpleAuth:
             raise self._Bad(f"Token type {self._token} requires a carrier")
         # name of token for cookie or param, Authentication scheme, or other header
         default_name: Optional[str] = \
-            "auth" if self._carrier in ("param", "cookie") else \
+            "AUTH" if self._carrier == "param" else \
+            "auth" if self._carrier == "cookie" else \
             "Bearer" if self._carrier == "bearer" else \
             "Auth" if self._carrier == "header" else \
             None
         self._name = conf.get("FSA_TOKEN_NAME", default_name)
         if need_carrier and not self._name:
             raise self._Bad(f"Token carrier {self._carrier} requires a name")
+        if self._carrier == "param":
+            self._names.add(self._name)
         # token realm…
         realm = conf.get("FSA_REALM", self._app.name)
         if self._token == "fsa":  # simplify realm for fsa
@@ -581,14 +614,15 @@ class FlaskSimpleAuth:
         #
         if "fake" not in self._auth and "FSA_FAKE_LOGIN" in conf:
             log.warning("ignoring directive FSA_FAKE_LOGIN")
-        self._login = conf.get("FSA_FAKE_LOGIN", "LOGIN")
         if "param" not in self._auth and "password" not in self._auth:
             if "FSA_PARAM_USER" in conf:
                 log.warning("ignoring directive FSA_PARAM_USER")
             if "FSA_PARAM_PASS" in conf:
                 log.warning("ignoring directive FSA_PARAM_PASS")
+        self._login = conf.get("FSA_FAKE_LOGIN", "LOGIN")
         self._userp = conf.get("FSA_PARAM_USER", "USER")
         self._passp = conf.get("FSA_PARAM_PASS", "PASS")
+        self._names.update([self._login, self._userp, self._passp])
         #
         # authentication and authorization hooks
         #
@@ -770,7 +804,7 @@ class FlaskSimpleAuth:
         assert self._http_auth
         auth = self._http_auth.get_auth()
         password = self._http_auth.get_auth_password(auth) \
-            if "http-token" not in self._auth else None
+            if "http-token" not in self._local.auth else None
         try:
             # NOTE "authenticate" signature is not very clean…
             user = self._http_auth.authenticate(auth, password)
@@ -1010,17 +1044,17 @@ class FlaskSimpleAuth:
         """Authenticate user or throw exception."""
 
         # safe because _user is reset before requests
-        if self._user_set:
-            return self._user
+        if self._local.user_set:
+            return self._local.user
 
         assert self._initialized, "FlaskSimpleAuth must be initialized"
 
         # try authentication schemes
         lae = None
-        for a in self._auth:
+        for a in self._local.auth:
             try:
-                self._user = self._FSA_AUTH[a](self)
-                if self._user:
+                self._local.user = self._FSA_AUTH[a](self)
+                if self._local.user:
                     break
             except FSAException as e:
                 lae = e
@@ -1028,13 +1062,13 @@ class FlaskSimpleAuth:
                 log.error(f"internal error in {a} authentication: {e}")
 
         # even if not set, we say that the answer is the right one.
-        self._user_set = True
+        self._local.user_set = True
 
         # rethrow last auth exception on failure
-        if required and not self._user:
+        if required and not self._local.user:
             raise lae or FSAException("missing authentication", 401)
 
-        return self._user
+        return self._local.user
 
     def current_user(self):
         """Return current authenticated user, if any."""
@@ -1110,19 +1144,18 @@ class FlaskSimpleAuth:
             def wrapper(*args, **kwargs):
 
                 # get user if needed
-                if not self._user_set:
+                if not self._local.user_set:
                     # possibly overwrite the authentication scheme
-                    # it will be restored in an after request hook
                     # NOTE this may or may not work because other settings may
                     #   not be compatible with the provided scheme…
                     if auth:
-                        self._saved_auth, self._auth = self._auth, auth
+                        self._local.auth = auth
                     try:
-                        self._user = self.get_user()
+                        self._local.user = self.get_user()
                     except FSAException as e:
                         return self._Res(e.message, e.status)
 
-                if not self._user:  # pragma no cover
+                if not self._local.user:  # pragma no cover
                     return self._Res("no auth", 401)
 
                 return self._safe_call(path, "authenticate", fun, *args, **kwargs)
@@ -1146,12 +1179,12 @@ class FlaskSimpleAuth:
             def wrapper(*args, **kwargs):
 
                 # track that some autorization check was performed
-                self._need_authorization = False
+                self._local.need_authorization = False
 
                 # check against all authorized groups/roles
                 for group in groups:
                     try:
-                        ok = self._user_in_group(self._user, group)
+                        ok = self._user_in_group(self._local.user, group)
                     except FSAException as fe:
                         return self._Res(fe.message, fe.status)
                     except Exception as e:
@@ -1175,15 +1208,18 @@ class FlaskSimpleAuth:
 
         def decorate(fun: Callable):
 
-            # for each parameter name, its expected type, cast and default value
+            # for each parameter name: type, cast, default value, http param name
             types: Dict[str, type] = {}
             typings: Dict[str, Callable[[str], Any]] = {}
             defaults: Dict[str, Any] = {}
+            names: Dict[str, str] = {}
 
             # parameters types/casts and defaults taken from signature
             sig, keywords = inspect.signature(fun), False
 
             for n, p in sig.parameters.items():
+                sn = n[1:] if n[0] == '_' and len(n) > 1 else n
+                names[n], names[sn] = sn, n
                 if n not in types and \
                    p.kind not in (p.VAR_KEYWORD, p.VAR_POSITIONAL):
                     # guess parameter type
@@ -1204,8 +1240,7 @@ class FlaskSimpleAuth:
 
                 for p, typing in typings.items():
                     # guess which function parameters are request parameters
-                    # parameter HTTP name
-                    pn = p[1:] if p[0] == '_' and len(p) > 1 else p
+                    pn = names[p]
                     if p not in kwargs:
                         # parameter p not yet encountered
                         if pn in params:
@@ -1236,6 +1271,10 @@ class FlaskSimpleAuth:
                     for p in params:
                         if p not in kwargs:
                             kwargs[p] = params[p]
+                elif self._debug:  # warn about unused parameters
+                    for p in params:
+                        if p not in names and p not in self._names:
+                            log.debug(f"unexpected parameter {p} on {path}")
 
                 return self._safe_call(path, "parameters", fun, *args, **kwargs)
 
@@ -1275,13 +1314,13 @@ class FlaskSimpleAuth:
             def wrapper(*args, **kwargs):
 
                 # track that some autorization check was performed
-                self._need_authorization = False
+                self._local.need_authorization = False
 
                 for domain, name, mode in perms:
                     val = kwargs[name]
 
                     try:
-                        ok = self._check_object_perms(self._user, domain, val, mode)
+                        ok = self._check_object_perms(self._local.user, domain, val, mode)
                     except FSAException as e:
                         return self._Res(e.message, e.status)
                     except Exception as e:
@@ -1312,7 +1351,7 @@ class FlaskSimpleAuth:
 
             @functools.wraps(fun)
             def wrapper(*args, **kwargs):
-                self._need_authorization = False
+                self._local.need_authorization = False
                 return self._safe_call(path, "no authorization", fun, *args, **kwargs)
 
             return wrapper
