@@ -367,10 +367,10 @@ def test_authorize():
     @app._fsa._group_auth("stuff", App.ADMIN)
     def stuff():
         return Response("", 200)
-    app._fsa._user = "dad"
+    app._fsa._local.user = "dad"
     res = stuff()
     assert res.status_code == 200
-    app._fsa._user = "hobbes"
+    app._fsa._local.user = "hobbes"
     res = stuff()
     assert res.status_code == 403
     try:
@@ -399,9 +399,9 @@ def test_self_care(client):
 def test_typed_params(client):
     res = check(200, client.get("/add/2", data={"a":"2.0", "b":"4.0"}))
     assert float(res.data) == 12.0
-    res = check(200, client.get("/mul/2", data={"j":"3", "k":"4"}))
+    res = check(200, client.get("/mul/2", data={"j":"3", "k":"4", "unused":"x"}))
     assert int(res.data) == 24
-    res = check(200, client.get("/mul/2", json={"j":5, "k":"4"}))
+    res = check(200, client.get("/mul/2", json={"j":5, "k":"4", "unused":"y"}))
     assert int(res.data) == 40
     check(400, client.get("/mul/1", data={"j":"3"}))
     check(400, client.get("/mul/1", data={"k":"4"}))
@@ -663,6 +663,35 @@ def test_reference():
     assert r3 == "1" and r3 != "one"
     assert r3.isdigit()
     assert repr("1") == repr(r3)
+    # thread local stuff
+    def gen_data(i):
+        return f"data: {i}"
+    r = fsa.Reference(fun=gen_data)
+    assert r._nthreads == 0
+    assert isinstance(r.__hash__(), int)
+    assert r._nthreads == 1
+    assert isinstance(r._local.obj, str)
+    assert r == "data: 0"
+    # another thread
+    import threading
+    def run(i):
+        assert r == f"data: {i}"
+    t1 = threading.Thread(target=run, args=(1,))
+    t1.start()
+    t1.join()
+    assert r._nthreads == 2
+    t2 = threading.Thread(target=run, args=(2,))
+    t2.start()
+    t2.join()
+    assert r._nthreads == 3
+    # local one is still ok
+    assert r == "data: 0"
+    # error
+    try:
+        r = fsa.Reference(obj="hello", fun=gen_data)
+        assert False, "should have raised an exception"
+    except Exception as e:
+        assert "reference cannot have both obj and fun" in str(e)
 
 def test_www_authenticate(client):
     push_auth(app._fsa, "param")
@@ -1145,6 +1174,7 @@ def test_warnings_and_errors():
         FSA_SUCH_DIRECTIVE="no-such-directive",
         FSA_AUTH=["basic", "token"],
         FSA_TOKEN_TYPE="fsa",
+        FSA_TOKEN_CARRIER="param",
         FSA_TOKEN_SIGN="signature-not-used-for-fsa-tokens",
         FSA_FAKE_LOGIN="not-used-if-no-fake",
         FSA_PARAM_USER="not-used-if-no-param",
